@@ -1,5 +1,7 @@
+import warnings
 import numpy as np
 import pandas as pd
+from datetime import datetime, timedelta
 from scipy.interpolate import interp1d
 from copy import deepcopy
 from mikeio.eum import ItemInfo
@@ -81,6 +83,26 @@ class Dataset:
 
     def __init__(self, data, time, items):
 
+        if isinstance(time, str):
+            # default single-step time
+            time = self.create_time(time)
+
+        if isinstance(items, int):
+            # default Undefined items
+            n_items = items
+            items = []
+            for j in range(n_items):
+                items.append(ItemInfo(f"Item {j+1}"))
+
+        if np.isscalar(data):
+            # create empty dataset
+            n_elements = data
+            n_items = len(items)
+            n_timesteps = len(time)
+            data = self.create_empty_data(
+                n_items=n_items, n_timesteps=n_timesteps, n_elements=n_elements
+            )
+
         n_items = len(data)
         n_timesteps = data[0].shape[0]
 
@@ -161,6 +183,19 @@ class Dataset:
 
         return Dataset(data, time, items)
 
+    def dropna(self):
+        "Remove time steps where all items are NaN"
+
+        # TODO consider all items
+        x = self[0]
+
+        # this seems overly complicated...
+        axes = tuple(range(1, x.ndim))
+        idx = np.where(~np.isnan(x).all(axis=axes))
+        idx = list(idx[0])
+
+        return self.isel(idx, axis=0)
+
     def flipud(self):
         "Flip dataset updside down"
 
@@ -543,6 +578,95 @@ class Dataset:
     def _ipython_key_completions_(self):
         return [x.name for x in self.items]
 
+    @staticmethod
+    def create_empty_data(n_items=1, n_timesteps=1, n_elements=None, shape=None):
+        data = []
+        if shape is None:
+            if n_elements is None:
+                raise ValueError("n_elements and shape cannot both be None")
+            else:
+                shape = n_elements
+        if np.isscalar(shape):
+            shape = [shape]
+        dati = np.empty(shape=(n_timesteps, *shape))
+        dati[:] = np.nan
+        for _ in range(n_items):
+            data.append(dati)
+        return data
+
+    @staticmethod
+    def create_time(start_time=None, dt=None, n_timesteps=None, end_time=None):
+        """create a equidistant time axis (calendar axis)
+
+        Parameters
+        ----------
+        start_time : datetime or str, optional
+            start_time, by default None (1970-1-1 00:00:00)
+            can optionally contain end_time e.g. '2018-01-01, 2018-02-01'
+        dt : float, optional
+            time step in seconds, by default None
+        n_timesteps : int, optional
+            number of timesteps, by default 1
+        end_time : datetime or str, optional
+            end_time, by default 3600s or deduced from other parameters
+
+        Returns
+        -------
+        pandas.DatetimeIndex
+            time axis which can be used to create new Dataset
+
+        Examples
+        ------
+        >>> t = Dateset.create_time('2018-1-1,2018-2-1', dt=1800)
+        >>> t = Dateset.create_time('2018-1-1', dt=1800, n_timesteps=48)
+        >>> t = Dateset.create_time('2018', dt=7200, end_time='2019')
+        """
+        if isinstance(start_time, str):
+            parts = start_time.split(",")
+            if len(parts) == 2:
+                end_time = parts[1]
+            start_time = pd.to_datetime(parts[0])
+        if isinstance(end_time, str):
+            end_time = pd.to_datetime(end_time)
+
+        if start_time is None:
+            # start_time = datetime.now()
+            start_time = datetime(1970, 1, 1, 0, 0, 0)
+
+        if dt is None:
+            if (end_time is not None) and (n_timesteps is not None):
+                dur = (end_time - start_time).total_seconds()
+                dt = 0.0
+                if (dur > 0) and (n_timesteps > 1):
+                    dt = dur / (n_timesteps - 1)
+            else:
+                warnings.warn("Too little information. Assuming dt=3600s.")
+                dt = 3600
+
+        if (end_time is None) and (n_timesteps is None):
+            warnings.warn("Too little information. Assuming n_timesteps=1.")
+            n_timesteps = 1
+
+        # find end time
+        if end_time is None:
+            tot_seconds = (n_timesteps - 1) * dt
+            end_time = start_time + timedelta(seconds=tot_seconds)
+        elif (end_time is not None) and (n_timesteps is not None):
+            # both are given, do they match?
+            tot_seconds = (n_timesteps - 1) * dt
+            end_time_2 = start_time + timedelta(seconds=tot_seconds)
+            if end_time != end_time_2:
+                raise ValueError("All parameters where given, but they do not match")
+
+        if dt < 0:
+            raise ValueError("dt cannot be negative")
+
+        if (end_time - start_time).total_seconds() < 0:
+            raise ValueError("end_time must be greater than start_time")
+
+        offset = pd.tseries.offsets.DateOffset(seconds=dt)
+        return pd.date_range(start=start_time, end=end_time, freq=offset)
+
     @property
     def is_equidistant(self):
         """Is Dataset equidistant in time?
@@ -552,6 +676,29 @@ class Dataset:
 
         return self.time.freq is not None
 
+    @property
+    def start_time(self):
+        """First time instance (as datetime)
+        """
+        return self.time[0].to_pydatetime()
+
+    @property
+    def end_time(self):
+        """Last time instance (as datetime)
+        """
+        return self.time[-1].to_pydatetime()
+
+    @property
+    def timestep(self):
+        """Time step in seconds if equidistant (and at 
+        least two time instances); otherwise None
+        """
+        dt = None
+        if len(self.time) > 1:
+            if self.is_equidistant:
+                dt = (self.time[1] - self.time[0]).total_seconds()
+        return dt
+
     @property
     def n_timesteps(self):
         """Number of time steps
